<?php
/**
 * The MIMP_Folder:: class provides a set of methods for dealing with
 * folders, accounting for subscription, errors, etc.
 *
 * $Horde: mimp/lib/Folder.php,v 1.21 2005/10/09 17:57:11 slusarz Exp $
 *
 * Copyright 2000-2005 Chuck Hagenbuch <chuck@horde.org>
 * Copyright 2000-2005 Jon Parise <jon@csh.rit.edu>
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 *
 * @author  Chuck Hagenbuch <chuck@horde.org>
 * @author  Jon Parise <jon@csh.rit.edu>
 * @package MIMP
 */
class MIMP_Folder {

    /**
     * Keep around identical lists so that we don't hit the server more that
     * once in the same page for the same thing.
     *
     * @var array
     */
    var $_listCache = array();

    /**
     * Keep track of mailbox names that we have complained about to prevent
     * giving the user identical error messages.
     *
     * @var array
     */
    var $_errorCache = array();

    /**
     * Cached results from the exists() function.
     *
     * @var array
     */
    var $_existsResults = array();

    /**
     * Returns a reference to the global MIMP_Folder object, only
     * creating it if it doesn't already exist. This ensures that only
     * one MIMP_Folder instance is instantiated for any given session
     *
     * This method must be invoked as:
     *   $mimp_folder = &MIMP_Folder::singleton();
     *
     * @return MIMP_Folder  The MIMP_Folder instance.
     */
    function &singleton()
    {
        static $folder;

        if (!isset($folder)) {
            $folder = new MIMP_Folder();
        }

        return $folder;
    }

    /**
     * Create a new IMAP folder if it does not already exist, and
     * subcribe to it as well if requested.
     *
     * @param string $folder      The full utf encoded folder to be created.
     * @param boolean $subscribe  A boolean describing whether or not to use
     *                            folder subscriptions.
     *
     * @return boolean  Whether or not the folder was successfully created.
     */
    function create($folder, $subscribe)
    {
        global $conf, $notification;

        /* Make sure we are not trying to create a duplicate folder */
        if ($this->exists($folder)) {
            $notification->push(sprintf(_("The folder \"%s\" already exists"), MIMP::displayFolder($folder)), 'horde.warning');
            return false;
        }

        /* Attempt to create the mailbox */
        if (!imap_createmailbox($_SESSION['mimp']['stream'], MIMP::serverString() . $folder)) {
            $notification->push(sprintf(_("The folder \"%s\" was not created. This is what the server said"), MIMP::displayFolder($folder)) .
                            ': ' . imap_last_error(), 'horde.error');
            return false;
        }

        /* If the user uses SUBSCRIBE, then add to the subscribe list */
        if ($subscribe &&
            !imap_subscribe($_SESSION['mimp']['stream'], MIMP::serverString() . $folder)) {
            $notification->push(sprintf(_("The folder \"%s\" was created but you were not subscribed to it."), MIMP::displayFolder($folder)), 'horde.warning');
        } else {
            /* The folder creation has been successful. */
            $notification->push(sprintf(_("The folder \"%s\" was successfully created."), MIMP::displayFolder($folder)), 'horde.success');
        }

        /* Update the MIMP_Tree object. */
        require_once MIMP_BASE . '/lib/IMAP/Tree.php';
        $mimptree = &MIMP_Tree::singleton();
        if ($mimptree) {
            $mimptree->insert($folder);
        }

        return true;
    }

    /**
     * Find out if a specific folder exists or not.
     *
     * @param string $folder    The full utf encoded folder name to be checked.
     * @param boolean $recheck  Should we recheck the folder?
     *
     * @return boolean  Whether or not the folder exists.
     */
    function exists($folder, $recheck = false)
    {
        if ($recheck || !isset($this->_existsResults[$folder])) {
            $path = MIMP::serverString() . $folder;
            $res = imap_list($_SESSION['mimp']['stream'], '', $path);
            $this->_existsResults[$folder] = ($res ? in_array($path, $res) : false);
        }

        return $this->_existsResults[$folder];
    }

    /**
     * Lists folders.
     *
     * @param boolean $sub   Should we list only subscribed folders?
     * @param array $filter  An list of mailboxes that should be left out of
     *                       the list.
     *
     * @return array  An array of folders, where each array alement is an
     *                associative array containing three values: 'val', with
     *                entire folder name after the server specification;
     *                'label', with the full-length folder name meant for
     *                display and 'abbrev', containing a shortened (26
     *                characters max) label for display in situations where
     *                space is short.
     */
    function flist($sub = false, $filter = array())
    {
        global $conf, $notification;

        if ($_SESSION['mimp']['base_protocol'] == 'pop3') {
            return array('INBOX' => array('val' => 'INBOX', 'label' => _("Inbox"), 'abbrev' => _("Inbox")));
        }

        $list = array();

        /* Figure out which list command to use. */
        $listcmd = ($sub) ? 'imap_lsub' : 'imap_list';

        /* Compute values that will uniquely identify this list. */
        $full_signature = md5(serialize(array($listcmd, $filter)));

        if (isset($this->_listCache[$listcmd])) {
            $mailboxes = $this->_listCache[$listcmd];
        } else {
            $mailboxes = array();
            $server = MIMP::serverString();
            foreach (array_keys($_SESSION['mimp']['namespace']) as $val) {
                $ns_info = MIMP::getNamespace($val);
                $delimiter = ($ns_info && $ns_info['delimiter']) ? $ns_info['delimiter'] : '';
                $search_string = ($delimiter) ? '*' : '';
                $hboxes = $listcmd($_SESSION['mimp']['stream'], $server, $val . $search_string);
                if (is_array($hboxes)) {
                    /* Strip off the prefix only if we are dealing with a
                     * personal namespace. */
                    $personal_ns = ($ns_info && ($ns_info['type'] == 'personal'));
                    $pos = strpos($hboxes[0], '}') + 1;
                    $prefix_len = 0;
                    if ($personal_ns && $search_string) {
                        $prefix_len = strlen($val);
                    }
                    foreach ($hboxes as $val2) {
                        $label = $stripped = substr($val2, $pos);
                        if ($prefix_len) {
                            $label = substr($label, $prefix_len);
                        }
                        if (strcasecmp('INBOX', $label) == 0) {
                            continue;
                        }
                        $mailboxes[$label] = array('v' => $stripped, 'd' => $delimiter);
                    }
                }
            }
            $this->_listCache[$listcmd] = $mailboxes;
        }

        if (!empty($mailboxes)) {
            $seen = array();

            require_once 'Horde/IMAP/Sort.php';
            // TODO: Fix use of delimiter here
            $delimiter = reset($_SESSION['mimp']['namespace']);
            $imap_mailboxes = &new IMAP_Sort($delimiter['delimiter']);
            $imap_mailboxes->sortMailboxesByKey($mailboxes);

            foreach ($mailboxes as $label => $mailbox) {
                if (empty($seen[$label])) {
                    $seen[$label] = true;

                    if (!($decoded_mailbox = String::convertCharset($mailbox['v'], 'UTF7-IMAP'))) {
                        if (empty($this->_errorCache[$mailbox['v']])) {
                            $notification->push(sprintf(_("The folder \"%s\" contains illegal characters in its name. It may cause problems. Please see your system administrator."), $label), 'horde.warning');
                            $this->_errorCache[$mailbox['v']] = true;
                        }
                    }

                    $parts = explode($mailbox['d']['delimiter'], $label);
                    $subfolder = array_pop($parts);
                    $parentparts = $parts;
                    $pstack = array();

                    while ($i = count($parentparts)) {
                        $parent = implode($mailbox['d']['delimiter'], $parentparts);
                        if (!empty($parent) &&
                            !isset($seen[$parent])) {
                            $seen[$parent] = true;
                            $plabel = str_repeat(' ', 4 * ($i - 1)) . String::convertCharset($parts[($i - 1)], 'UTF7-IMAP');
                            if (strlen($plabel) > 26) {
                                $pabbrev = String::substr($plabel, 0, 10) . '...' . String::substr($plabel, -13, 13);
                            } else {
                                $pabbrev = $plabel;
                            }
                            $pstack[$parent] = array('val' => '', 'label' => $plabel, 'abbrev' => $pabbrev);
                        }
                        array_pop($parentparts);
                    }

                    if (count($pstack)) {
                        $pstack = array_reverse($pstack);
                        foreach ($pstack as $pkey => $pval) {
                            $list[$pkey] = $pval;
                        }
                    }

                    $folded = str_repeat(' ', 4 * count($parts)) . String::convertCharset($subfolder, 'UTF7-IMAP');

                    if (strlen($folded) > 26) {
                        $abbrev = String::substr($folded, 0, 10) . '...' . String::substr($folded, -13, 13);
                    } else {
                        $abbrev = $folded;
                    }

                    if (($_SESSION['mimp']['dotfiles'] || (strlen($label) > 0 && (($label[0] != '.') && !strstr($label, $mailbox['d'] . '.'))))
                        && !in_array($mailbox['v'], $filter)) {
                        $list[$mailbox['v']] = array('val' => $mailbox['v'], 'label' => $folded, 'abbrev' => $abbrev);
                    }
                }
            }
        }

        /* Add the INBOX on top of list if not in the filter list. */
        if (!in_array('INBOX', $filter)) {
            $list = array('INBOX' => array('val' => 'INBOX', 'label' => _("Inbox"), 'abbrev' => _("Inbox"))) + $list;
        }

        return $list;
    }

    /**
     * Returns an array of folders. This is a wrapper around the flist()
     * function which reduces the number of arguments needed if we can assume
     * that MIMP's full environment is present.
     *
     * @param array $filter  An array of mailboxes to ignore.
     * @param boolean $sub   If set, will be used to determine if we should
     *                       list only subscribed folders.
     *
     * @return array  The array of mailboxes returned by flist().
     */
    function flist_MIMP($filter = array(), $sub = null)
    {
        return $this->flist(is_null($sub) ? $GLOBALS['prefs']->getValue('subscribe') : $sub, $filter);
    }

}
